open Printf
open Code

exception PasFini

(*****************************************)
(* Type des blocs de base, usage interne *)
(*****************************************)

type basic_block = {enter:Gen.label ; mutable succ:stm ; body:stm list}

(*********************************************)
(* Utilitaires d'affichage du graphe de flot *)
(*********************************************)

(* Fonction pratique qui donne les tiquettes des successeurs d'un bloc *)
let get_succ b = match b.succ with
| Jump lab -> [lab]
| Cjump (_,_,_,lab1,lab2) -> [lab1 ; lab2]
|  _ -> assert false


let print_flow_graph blocks =
  let print_node_and_arcs block =
    let lab = block.enter in
    begin match block with
    | {body=[] ; succ=Jump _} ->
        printf "  %s [fillcolor=gray, style=filled];\n"
          (Gen.label_string lab)
    | _ ->  ()
    end ;
    let print_arc tolab =
      printf "  %s -> %s;\n"
        (Gen.label_string lab)
        (Gen.label_string tolab)
    in
    List.iter print_arc (get_succ block) 
  in
  List.iter print_node_and_arcs blocks

(*******************************)
(* Fabriquer les blocs de base *)
(*******************************)

(* Dcoupe en blocs de base.
  Type:
    Gen.label -> Gen.label -> code -> basic_block list
  Appel: 
     cut_blocks lstart lend code 
   Les arguments sont
     - lstart tiquette d'entre dasn la fonction.
     - lexit tiquette de sortie (ie celle de l'pilogue).
     - code: liste des instructions du corps de la fonction.
   Revoie la liste des blocs de base, avec
     - Le premier bloc porte l'tiquette lstart.
     - Un saut vers lend signifie retourner de la fonction.
*)


let cut_blocks start_label exit_label code = raise PasFini

let code_to_blocks f code =
  let start_label = Frame.frame_name f in
  let exit_label = Frame.frame_return f in
  cut_blocks start_label exit_label code;;


(***************************************************)
(* Reconstituer le code  partir des blocs de base *)
(***************************************************)

let blocks_to_code blocks =  raise PasFini

(************************************)
(* Optimisation  trou de serrure  *)
(************************************)

(*
   Faire le minimum, c'est--dire forcer
   Cjump (_,_,_,l1,l2) ; Label l2 ; ..
*)
let neg = function
  | Req -> Rne | Rne -> Req
  | Rle -> Rgt | Rge -> Rlt | Rlt -> Rge | Rgt -> Rle

let rec peephole f = function
  | Cjump (binop, c1, c2, l1, l2) :: t ->
      begin match t with
      | Label l3 :: t3 when l1 = l3 ->
          Cjump (neg binop, c1, c2, l2, l1) ::
          Label l3 :: peephole f t3
      | Label l3 :: t3 when l2 = l3 ->
          Cjump (binop, c1, c2, l1, l2) ::
          Label l3 :: peephole f t3
      | _ ->
          let l4 = Gen.new_label () in
          Cjump (binop, c1, c2, l1, l4) ::
          Label l4 :: Jump l2 :: peephole f t
      end
  | h :: t -> h :: peephole f t 
  | [] -> []
              

(* Optimisation du contrle de f (dont le code est c) *)
let opt_trace f c = 
  let blocks = code_to_blocks f c in
  if !Misc.dump_flowgraph then print_flow_graph blocks;
  peephole f (blocks_to_code blocks)

(* Idem, mais inoffensif *)
let opt_trace f c = peephole f c

(**************************************************)
(* Affichage du code intermdiaire  divers stades *)
(**************************************************)

let pcode pfun f c = if !Misc.dump_icode then pfun f c

let porig f c =
  printf "*** Code gnr ***\n" ;
  Iprint.frame_code (f,c)


let pcanon _ c =
  printf "*** Code canonique et linaris ***\n" ;
  Iprint.code c

let popt _ c =
  printf "*** Code aprs optimisation du contrle ***\n" ;
  Iprint.code c ;
  print_newline ()

(*********************************************)    
(* Itrer sur chaque fonction d'un programme *)
(*********************************************)    

open Trans

let map f p = 
  { number_of_globals = p.number_of_globals ;
    main =
      (match p.main with (frame,code) -> frame,f frame code);
    procedures =
    List.map (fun (frame, code) -> frame, f frame code)
      p.procedures;
  } 


let program p =
  if !Misc.dump_flowgraph then
    printf "digraph \"%s\" {\n" (Filename.basename !Misc.prog);
  let output =
    map (fun f c ->
      pcode porig f c ;
      let c = Canon.code c in
      pcode pcanon f c ;
      let c = opt_trace f c in
      pcode popt f c ;
      c) p in
  if !Misc.dump_flowgraph then printf "}\n";
  output
